'use strict';

const axios = require("axios"),
    Q = require("q"),
    DigestClient = require('digest-fetch');


const MappedErrorCodes = {
    Unknown: "Unknown",
    NoResponse: "NoResponse",
    Refused: "Refused",
    HttpError: "HttpError",
    StreamError: "StreamError",
    AuthNotSupported: "AuthNotSupported"
}

module.exports = {
    /**
     * Starts to read the stream the "success" callback is called on every frame
     * @param success
     * @param error
     * @param url
     * @param user
     * @param password
     * @param errorCount
     */
    startStream(success, error, url, user, password, errorCount = 0) {
        let _errorCount = errorCount;
        this._isRemoteConnectUrl = url.indexOf("dyndns.loxonecloud") > -1;

        this._successCallback = b64String => {
            _errorCount = 0;
            if (this._readyForFrame) {
                this._readyForFrame = false;
                success(b64String);
            }
        };
        this._errorCallback = (errorUrl, err) => {
            _errorCount++;
            if (_errorCount >= 3 || err.message === 'canceled' || err.type === "aborted") {
                this.stopStream();
                if (err.type === "aborted" || err.message === 'canceled') {
                    // nothing to do the user canceled the stream
                } else {
                    error(this._parseError(err));
                }
            } else {
                this.startStream(success, error, url, user, password, _errorCount);
            }
        };

        this._getAuthHeader({
            url: url,
            user: user,
            pwd: password
        }).then(isDigest => {
            this.stopStream();
            this._abortHandler = new AbortController();
            const isBasic = !isDigest && (!!user || !!password);
            const client = new DigestClient(user, password, {
                basic: isBasic,
            });

            this._fetchImage({client, user, password, url, isDigest})
        }, err => {
            this._errorCallback(url, err);
        });
    },


    _fetchImage({client, user, password, url, isDigest}) {
        const options = {
            signal: this._abortHandler.signal,
        }

        if (isDigest) {
            options.adoptUri = (url) => {
                if (this._isRemoteConnectUrl) {
                    const regex =/(\/proxy\/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{16})?(\/?.*)/i;
                    const matches = regex.exec(url);

                    return matches[2]
                } else {
                    return url;
                }
            }
        }

        client.fetch(url, options).then(resp => {
            if (resp.status === 200) {
                const contentType = resp.headers.raw()['content-type'][0]

                if (contentType.startsWith("multipart/x-mixed-replace")) {
                    let chunks = [],
                        soi = Buffer.from([0xff, 0xd8]), // Start Of Input
                        eoi = Buffer.from([0xff, 0xd9]); // End Of Input

                    resp.body.on("data", chunk => {
                        if (chunks.length && chunk.indexOf(eoi) !== -1) {
                            chunks.push(chunk.slice(0, chunk.indexOf(eoi) + 1 + 1)); // +1 is for the second bit of the eoi buffer, the other +1 is for get the length instead of the index
                            this._successCallback(Buffer.concat(chunks).toString("base64"));
                        }
                        if (chunk.indexOf(soi) !== -1) { // Check if the data chunk contains the JPEG start magic
                            chunks = [];
                            chunks.push(chunk.slice(chunk.indexOf(soi), chunk.length));
                        } else if (chunks.length) {
                            chunks.push(chunk);
                        }
                    });
                } else if (contentType.includes("image/png") || contentType.includes("image/jpeg")) {
                    resp.arrayBuffer().then((chunk) => {
                        this._successCallback(Buffer.from(chunk).toString('base64'));

                        if (!this._abortHandler.signal.aborted) {
                            this._fetchImage({client, url, user, password, isDigest})
                        }
                    }, (e) => {
                        this._errorCallback(url, e);
                    });
                } else {
                    this._errorCallback(url, `Unsupported contentType: ${contentType}`);
                }
            } else {
                this._errorCallback(url, {
                    response: {
                        status: resp.status,
                        statusText: resp.statusText
                    }
                });
            }
        }, err => {
            this._errorCallback(url, err);
        })
    },

    _parseError(err) {
        let errorObj = {};

        if (err) {
            if (err.code) {
                switch (err.code) {
                    case "ENOTFOUND":
                        errorObj.errorId = MappedErrorCodes.NoResponse;
                        errorObj.errorText = err.message;
                        break;
                    case "ECONNREFUSED":
                        errorObj.errorId = MappedErrorCodes.Refused;
                        errorObj.errorText = err.message;
                        break;
                    case "ETIMEDOUT":
                        errorObj.errorId = MappedErrorCodes.NoResponse;
                        errorObj.errorText = err.message;
                        break;
                    case "ECONNRESET":
                    case "AUTH_ERROR":
                        errorObj.errorId = MappedErrorCodes.StreamError;
                        errorObj.errorText = err.message;
                        break;
                    case "AUTH_UNSUPPORTED":
                        errorObj.errorId = MappedErrorCodes.AuthNotSupported;
                        errorObj.errorText = err.message;
                        break;
                    default:
                        errorObj.errorId = MappedErrorCodes.Unknown;
                        break;
                }
            } else if (err.response.status) {
                // eg. forbidden
                errorObj.errorId = MappedErrorCodes.HttpError;
                errorObj.errorCode = err.response.status;
                errorObj.errorText = err.response.statusText;
            }

        } else {
            errorObj.errorId = MappedErrorCodes.Unknown;

        }

        return JSON.stringify(errorObj);
    },

    _getAuthHeader({url}) {
        let abortController = new AbortController(),
            authValue = null,
            resolveDef = Q.defer(),
            basicAuthValue;

        axios.get(url, {
            responseType: "stream",
            signal: abortController.signal
        }).then(resp => {
            abortController.abort();
            abortController.signal.onabort = () => {
                resp.data.req.abort();
                resp.data.req.destroy();
            };

            if (resp.headers && resp.headers["content-type"] && resp.headers["content-type"].startsWith("multipart/x-mixed-replace")) {
                this.stopStream();
                this._abortHandler = new AbortController();
                let chunks = [],
                    soi = Buffer.from([0xff, 0xd8]), // Start Of Input
                    eoi = Buffer.from([0xff, 0xd9]); // End Of Input


                resp.data.on("data", chunk => {
                    // since the introduction from the remote connect, the chunk we get can contain the "End of Input" and also the "Start of Input",
                    // so we first have to check if the chunk contains the "End of Input"
                    if (chunks.length && chunk.indexOf(eoi) !== -1) {
                        chunks.push(chunk.slice(0, chunk.indexOf(eoi) + 1 + 1)); // +1 is for the second bit of the eoi buffer, the other +1 is for get the length instead of the index
                        this._successCallback(Buffer.concat(chunks).toString("base64"));
                    }
                    if (chunk.indexOf(soi) !== -1) { // Check if the data chunk contains the JPEG start magic
                        chunks = [];
                        chunks.push(chunk.slice(chunk.indexOf(soi), chunk.length));
                    } else if (chunks.length) {
                        chunks.push(chunk);
                    }
                })

                this._abortHandler.signal.onabort = () => {
                    resp.data.req.abort();
                    resp.data.req.destroy();
                };
            } else {
                resolveDef.resolve(authValue);
            }
        }, err => {
            if (err.response) {
                let { headers } = err.response;

                if (headers && headers["www-authenticate"]) {
                    let authMethods = headers["www-authenticate"].split(/(Basic|Digest)/),
                        digestAuthHeaderIdx,
                        basicAuthHeaderIdx;

                    digestAuthHeaderIdx = authMethods.findIndex(string => {
                        return string === "Digest";
                    })

                    if (digestAuthHeaderIdx !== -1) {
                        return resolveDef.resolve(true)
                    }

                    basicAuthHeaderIdx = authMethods.findIndex(string => {
                        return string === "Basic";
                    })

                    if (basicAuthHeaderIdx !== -1) {
                        return resolveDef.resolve(false)
                    }

                    return resolveDef.reject({
                        code: "AUTH_UNSUPPORTED",
                        message: "Unsupported Authentication method provided"
                    })

                } else {
                    return resolveDef.reject({
                        code: "AUTH_ERROR",
                        message: "No Authentication method provided"
                    });
                }
            } else {
                return resolveDef.reject(err);
            }
        });

        return resolveDef.promise;
    },

    /**
     * Stops the currently active stream
     */
    stopStream() {
        this._abortHandler && this._abortHandler.abort();
        this.imageProcessed();
    },

    /**
     * Call this when the UI is ready processing the image,
     * no further frame will be emitted to the callback until this method is called
     */
    imageProcessed() {
        this._readyForFrame = true;
    },
};
